# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from collections import namedtuple

import tornado
import tornado.gen

from tchannel import retry

from ..glossary import DEFAULT_TIMEOUT
from ..messages import ErrorCode
from ..messages.common import FlagsType
from ..messages.common import StreamState
from ..serializer.raw import RawSerializer
from ..zipkin.trace import Trace
from .stream import InMemStream
from .util import get_arg


class Request(object):
    """Represents an incoming request to an endpoint.

    Request class is used to represent the CallRequestMessage at User's level.
    This is going to hide the protocol level message information.
    """

    # TODO deprecate in favor of top-level tchannel.Request
    # TODO decide which elements inside "message" object to expose to user.

    def __init__(
        self,
        id=None,
        flags=FlagsType.none,
        ttl=DEFAULT_TIMEOUT,
        tracing=None,
        service=None,
        headers=None,
        checksum=None,
        argstreams=None,
        serializer=None,
        endpoint=None,
    ):
        self.flags = flags
        self.ttl = ttl
        self.service = service
        self.tracing = tracing or Trace()
        # argstreams is a list of InMemStream/PipeStream objects
        self.argstreams = argstreams or [InMemStream(),
                                         InMemStream(),
                                         InMemStream()]
        self.checksum = checksum
        self.id = id
        self.headers = headers or {}
        self.state = StreamState.init
        self.serializer = serializer or RawSerializer()

        self.is_streaming_request = self._is_streaming_request()
        if not self.is_streaming_request:
            self._copy_argstreams = [
                self.argstreams[0].clone(),
                self.argstreams[1].clone(),
                self.argstreams[2].clone(),
            ]

        self.endpoint = endpoint or ""

    def rewind(self, id=None):
        self.id = id
        if not self.is_streaming_request:
            self.argstreams = [
                self._copy_argstreams[0].clone(),
                self._copy_argstreams[1].clone(),
                self._copy_argstreams[2].clone(),
            ]
        self.state = StreamState.init
        self.tracing = Trace()

    @property
    def arg_scheme(self):
        return self.headers.get('as', None)

    def set_exception(self, exception):
        for stream in self.argstreams:
            stream.set_exception(exception)
            stream.close()

    def close_argstreams(self, force=False):
        for stream in self.argstreams:
            if stream.auto_close or force:
                stream.close()

    @tornado.gen.coroutine
    def get_header(self):
        """Get the header value from the request.

        :return: a future contains the deserialized value of header
        """
        raw_header = yield get_arg(self, 1)
        if not self.serializer:
            raise tornado.gen.Return(raw_header)
        else:
            header = self.serializer.deserialize_header(raw_header)
            raise tornado.gen.Return(header)

    @tornado.gen.coroutine
    def get_body(self):
        """Get the body value from the resquest.

        :return: a future contains the deserialized value of body
        """

        raw_body = yield get_arg(self, 2)
        if not self.serializer:
            raise tornado.gen.Return(raw_body)
        else:
            body = self.serializer.deserialize_body(raw_body)
            raise tornado.gen.Return(body)

    def get_header_s(self):
        """Get the raw stream of header.

        :return: the argstream of header
        """
        return self.argstreams[1]

    def get_body_s(self):
        """Get the raw stream of body.

        :return: the argstream of body
        """
        return self.argstreams[2]

    def _is_streaming_request(self):
        """check request is stream request or not"""
        arg2 = self.argstreams[1]
        arg3 = self.argstreams[2]
        return not (isinstance(arg2, InMemStream) and
                    isinstance(arg3, InMemStream) and
                    ((arg2.auto_close and arg3.auto_close) or (
                        arg2.state == StreamState.completed and
                        arg3.state == StreamState.completed)))

    def should_retry_on_error(self, error):
        """rules for retry

        :param error:
            ProtocolException that returns from Server
        """

        if self.is_streaming_request:
            # not retry for streaming request
            return False

        retry_flag = self.headers.get('re', retry.DEFAULT)

        if retry_flag == retry.NEVER:
            return False

        if error.code in [ErrorCode.bad_request, ErrorCode.cancelled,
                          ErrorCode.unhealthy]:
            return False
        elif error.code in [ErrorCode.busy, ErrorCode.declined]:
            return True
        elif error.code is ErrorCode.timeout:
            return retry_flag is not retry.CONNECTION_ERROR
        elif error.code in [ErrorCode.network_error,
                            ErrorCode.fatal,
                            ErrorCode.unexpected]:
            return retry_flag is not retry.TIMEOUT
        else:
            return False


class TransportMetadata(
    namedtuple('_Metadata', 'flags ttl service id headers')
):
    """A read-only representation of the metadata contained in the Request."""

    @classmethod
    def from_request(cls, request):
        return cls(
            flags=request.flags,
            ttl=request.ttl,
            service=request.service,
            id=request.id,
            headers=request.headers,
        )



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#
# Autogenerated by Thrift Compiler (0.9.2)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,slots,dynamic
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException

from thrift.protocol.TBase import TBase, TExceptionBase


class AnnotationType(TBase):
  BOOL = 0
  BYTES = 1
  I16 = 2
  I32 = 3
  I64 = 4
  DOUBLE = 5
  STRING = 6

  _VALUES_TO_NAMES = {
    0: "BOOL",
    1: "BYTES",
    2: "I16",
    3: "I32",
    4: "I64",
    5: "DOUBLE",
    6: "STRING",
  }

  _NAMES_TO_VALUES = {
    "BOOL": 0,
    "BYTES": 1,
    "I16": 2,
    "I32": 3,
    "I64": 4,
    "DOUBLE": 5,
    "STRING": 6,
  }


class Endpoint(TBase):
  """
  Attributes:
   - ipv4
   - port
   - serviceName
  """

  __slots__ = [ 
    'ipv4',
    'port',
    'serviceName',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'ipv4', None, None, ), # 1
    (2, TType.I32, 'port', None, None, ), # 2
    (3, TType.STRING, 'serviceName', None, None, ), # 3
  )

  def __init__(self, ipv4=None, port=None, serviceName=None,):
    self.ipv4 = ipv4
    self.port = port
    self.serviceName = serviceName

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.ipv4)
    value = (value * 31) ^ hash(self.port)
    value = (value * 31) ^ hash(self.serviceName)
    return value


class Annotation(TBase):
  """
  Attributes:
   - timestamp
   - value
   - duration
  """

  __slots__ = [ 
    'timestamp',
    'value',
    'duration',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.DOUBLE, 'timestamp', None, None, ), # 1
    (2, TType.STRING, 'value', None, None, ), # 2
    (3, TType.I32, 'duration', None, None, ), # 3
  )

  def __init__(self, timestamp=None, value=None, duration=None,):
    self.timestamp = timestamp
    self.value = value
    self.duration = duration

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.timestamp)
    value = (value * 31) ^ hash(self.value)
    value = (value * 31) ^ hash(self.duration)
    return value


class BinaryAnnotation(TBase):
  """
  Attributes:
   - key
   - stringValue
   - doubleValue
   - boolValue
   - bytesValue
   - intValue
   - annotationType
  """

  __slots__ = [ 
    'key',
    'stringValue',
    'doubleValue',
    'boolValue',
    'bytesValue',
    'intValue',
    'annotationType',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'key', None, None, ), # 1
    (2, TType.STRING, 'stringValue', None, None, ), # 2
    (3, TType.DOUBLE, 'doubleValue', None, None, ), # 3
    (4, TType.BOOL, 'boolValue', None, None, ), # 4
    (5, TType.STRING, 'bytesValue', None, None, ), # 5
    (6, TType.I64, 'intValue', None, None, ), # 6
    (7, TType.I32, 'annotationType', None, None, ), # 7
  )

  def __init__(self, key=None, stringValue=None, doubleValue=None, boolValue=None, bytesValue=None, intValue=None, annotationType=None,):
    self.key = key
    self.stringValue = stringValue
    self.doubleValue = doubleValue
    self.boolValue = boolValue
    self.bytesValue = bytesValue
    self.intValue = intValue
    self.annotationType = annotationType

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.key)
    value = (value * 31) ^ hash(self.stringValue)
    value = (value * 31) ^ hash(self.doubleValue)
    value = (value * 31) ^ hash(self.boolValue)
    value = (value * 31) ^ hash(self.bytesValue)
    value = (value * 31) ^ hash(self.intValue)
    value = (value * 31) ^ hash(self.annotationType)
    return value


class Span(TBase):
  """
  Attributes:
   - traceId
   - host
   - name
   - id
   - parentId
   - annotations
   - binaryAnnotations
   - debug
  """

  __slots__ = [ 
    'traceId',
    'host',
    'name',
    'id',
    'parentId',
    'annotations',
    'binaryAnnotations',
    'debug',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'traceId', None, None, ), # 1
    (2, TType.STRUCT, 'host', (Endpoint, Endpoint.thrift_spec), None, ), # 2
    (3, TType.STRING, 'name', None, None, ), # 3
    (4, TType.STRING, 'id', None, None, ), # 4
    (5, TType.STRING, 'parentId', None, None, ), # 5
    (6, TType.LIST, 'annotations', (TType.STRUCT,(Annotation, Annotation.thrift_spec)), None, ), # 6
    (7, TType.LIST, 'binaryAnnotations', (TType.STRUCT,(BinaryAnnotation, BinaryAnnotation.thrift_spec)), None, ), # 7
    (8, TType.BOOL, 'debug', None, False, ), # 8
  )

  def __init__(self, traceId=None, host=None, name=None, id=None, parentId=None, annotations=None, binaryAnnotations=None, debug=thrift_spec[8][4],):
    self.traceId = traceId
    self.host = host
    self.name = name
    self.id = id
    self.parentId = parentId
    self.annotations = annotations
    self.binaryAnnotations = binaryAnnotations
    self.debug = debug

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.traceId)
    value = (value * 31) ^ hash(self.host)
    value = (value * 31) ^ hash(self.name)
    value = (value * 31) ^ hash(self.id)
    value = (value * 31) ^ hash(self.parentId)
    value = (value * 31) ^ hash(self.annotations)
    value = (value * 31) ^ hash(self.binaryAnnotations)
    value = (value * 31) ^ hash(self.debug)
    return value


class Response(TBase):
  """
  Attributes:
   - ok
  """

  __slots__ = [ 
    'ok',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.BOOL, 'ok', None, None, ), # 1
  )

  def __init__(self, ok=None,):
    self.ok = ok

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.ok)
    return value




# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from functools import wraps

import mock
import contextlib2
from tornado import gen

from tchannel import schemes
from tchannel.errors import TChannelError
from tchannel.tornado import TChannel
from tchannel.tornado.response import Response
from tchannel.tornado.stream import maybe_stream
from tchannel.tornado.stream import read_full

from .proxy import VCRProxy


_TChannel_request = TChannel.request


@contextlib2.contextmanager
def force_reset():
    with contextlib2.ExitStack() as exit_stack:
        exit_stack.enter_context(
            mock.patch.object(TChannel, 'request', _TChannel_request)
        )
        yield


class PatchedClientOperation(object):

    def __init__(
        self,
        vcr_client,
        original_tchannel,
        hostport=None,
        service=None,
        arg_scheme=None,
        retry=None,
        parent_tracing=None,
        score_threshold=None,
    ):
        self.vcr_client = vcr_client
        self.hostport = hostport
        self.service = service or ''
        self.arg_scheme = arg_scheme or schemes.DEFAULT
        self.original_tchannel = original_tchannel

        # TODO what to do with retry, parent_tracing and score_threshold

    @gen.coroutine
    def send(self, arg1, arg2, arg3,
             headers=None,
             traceflag=None,
             retry_limit=None,
             ttl=None):
        arg1, arg2, arg3 = map(maybe_stream, [arg1, arg2, arg3])

        endpoint = yield read_full(arg1)

        headers = headers or {}
        headers.setdefault('as', self.arg_scheme)

        vcr_request = VCRProxy.Request(
            serviceName=self.service.encode('utf-8'),
            hostPort=self.hostport,
            knownPeers=self.original_tchannel.peers.hosts,
            endpoint=endpoint,
            headers=(yield read_full(arg2)),
            body=(yield read_full(arg3)),
            argScheme=getattr(VCRProxy.ArgScheme, self.arg_scheme.upper()),
            transportHeaders=[
                VCRProxy.TransportHeader(k, v) for k, v in headers.items()
            ],
        )

        # TODO what to do with traceflag, attempt-times, ttl
        # TODO catch protocol errors

        with force_reset():
            vcr_response_future = self.vcr_client.send(vcr_request)
        try:
            vcr_response = yield vcr_response_future
        except VCRProxy.RemoteServiceError as e:
            raise TChannelError.from_code(
                e.code,
                description=(
                    "The remote service threw a protocol error: %s" %
                    e.message
                )
            )
        response = Response(
            code=vcr_response.code,
            argstreams=[
                maybe_stream(endpoint),
                maybe_stream(vcr_response.headers),
                maybe_stream(vcr_response.body),
            ],
            # TODO headers=vcr_response.transportHeaders,
        )

        raise gen.Return(response)


class Patcher(object):
    """Monkey patches classes to use a VCRProxyClient to send requests."""

    def __init__(self, vcr_client):
        """
        :param vcr_client:
            The VCRProxyClient through which requests will be made.
        """
        self.vcr_client = vcr_client
        self._exit_stack = contextlib2.ExitStack()

    def _patch_request(self):

        @wraps(_TChannel_request)
        def request(channel, *args, **kwargs):
            return PatchedClientOperation(
                self.vcr_client, channel, *args, **kwargs
            )

        return mock.patch.object(TChannel, 'request', request)

    def __enter__(self):
        self._exit_stack.enter_context(self._patch_request())

    def __exit__(self, *args):
        self._exit_stack.close()

    def __call__(self, function):
        # being used as a decorator

        @wraps(function)
        def new_function(*args, **kwargs):
            with self:
                return function(*args, **kwargs)

        return new_function



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import inspect
import types

from tchannel.errors import ValueExpectedError
from tchannel.errors import OneWayNotSupportedError
from tchannel.serializer.thrift import ThriftSerializer

from .reflection import get_service_methods, get_module_name


def thrift_request_builder(service, thrift_module, hostport=None,
                           thrift_class_name=None):
    """Provide TChannel compatibility with Thrift-generated modules.

    The service this creates is meant to be used with TChannel like so:

    .. code-block:: python

        from tchannel import TChannel, thrift_request_builder
        from some_other_service_thrift import some_other_service

        tchannel = TChannel('my-service')

        some_service = thrift_request_builder(
            service='some-other-service',
            thrift_module=some_other_service
        )

        resp = tchannel.thrift(
            some_service.fetchPotatoes()
        )

    :param string service:
        Name of Thrift service to call. This is used internally for
        grouping and stats, but also to route requests over Hyperbahn.

    :param thrift_module:
        The top-level module of the Apache Thrift generated code for
        the service that will be called.

    :param string hostport:
        When calling the Thrift service directly, and not over Hyperbahn,
        this 'host:port' value should be provided.

    :param string thrift_class_name:
        When the Apache Thrift generated Iface class name does not match
        thrift_module, then this should be provided.
    """

    # start with a request maker instance
    maker = ThriftRequestMaker(
        service=service,
        thrift_module=thrift_module,
        hostport=hostport,
        thrift_class_name=thrift_class_name
    )

    # create methods that mirror thrift client
    # and each return ThriftRequest
    methods = _create_methods(thrift_module)

    # then attach to instane
    for name, method in methods.iteritems():
        method = types.MethodType(method, maker, ThriftRequestMaker)
        setattr(maker, name, method)

    return maker


class ThriftRequestMaker(object):

    def __init__(self, service, thrift_module,
                 hostport=None, thrift_class_name=None):

        self.service = service
        self.thrift_module = thrift_module
        self.hostport = hostport

        if thrift_class_name is not None:
            self.thrift_class_name = thrift_class_name
        else:
            self.thrift_class_name = get_module_name(self.thrift_module)

    def _make_request(self, method_name, args, kwargs):

        result_type = self._get_result_type(method_name)

        if result_type is None:
            raise OneWayNotSupportedError(
                'TChannel+Thrift does not currently support oneway '
                'procedures.'
            )

        endpoint = self._get_endpoint(method_name)
        call_args = self._get_call_args(method_name, args, kwargs)

        request = ThriftRequest(
            service=self.service,
            endpoint=endpoint,
            result_type=result_type,
            call_args=call_args,
            hostport=self.hostport
        )

        return request

    def _get_endpoint(self, method_name):

        endpoint = '%s::%s' % (self.thrift_class_name, method_name)

        return endpoint

    def _get_args_type(self, method_name):

        args_type = getattr(self.thrift_module, method_name + '_args')

        return args_type

    def _get_result_type(self, method_name):

        # if None then result_type is oneway
        result_type = getattr(
            self.thrift_module, method_name + '_result', None
        )

        return result_type

    def _get_call_args(self, method_name, args, kwargs):

        args_type = self._get_args_type(method_name)

        params = inspect.getcallargs(
            getattr(self.thrift_module.Iface, method_name),
            self,
            *args,
            **kwargs
        )
        params.pop('self')  # self is already known

        call_args = args_type()
        for name, value in params.items():
            setattr(call_args, name, value)

        return call_args


class ThriftRequest(object):

    __slots__ = (
        'service', 'endpoint', 'result_type', 'call_args', 'hostport',
        '_serializer',
    )

    # TODO - implement __repr__

    def __init__(self, service, endpoint, result_type,
                 call_args, hostport=None, serializer=None):

        self.service = service
        self.endpoint = endpoint
        self.result_type = result_type
        self.call_args = call_args
        self.hostport = hostport

        if not serializer:
            serializer = ThriftSerializer(self.result_type)
        self._serializer = serializer

    def get_serializer(self):
        return self._serializer

    def read_body(self, body):
        """Handles the response body for this request.

        If the response body includes a result, returns the result unwrapped
        from the response union. If the response contains an exception, raises
        that exception.
        """
        result_spec = self.result_type.thrift_spec

        # raise application exception, if present
        for exc_spec in result_spec[1:]:
            exc = getattr(body, exc_spec[2])
            if exc is not None:
                raise exc

        # success - non-void
        if len(result_spec) >= 1 and result_spec[0] is not None:

            # value expected, but got none
            # TODO - server side should use this same logic
            if body.success is None:
                raise ValueExpectedError(
                    'Expected a value to be returned for %s, '
                    'but recieved None - only void procedures can '
                    'return None.' % self.endpoint
                )

            return body.success

        # success - void
        else:
            return None


def _create_methods(thrift_module):

    # TODO - this method isn't needed, instead, do:
    #
    # for name in get_service_methods(...):
    #   method = _create_method(...)
    #     # ...
    #

    methods = {}
    method_names = get_service_methods(thrift_module.Iface)

    for method_name in method_names:

        method = _create_method(method_name)
        methods[method_name] = method

    return methods


def _create_method(method_name):

    # TODO - copy over entire signature using @functools.wraps(that_function)
    # or wrapt on Iface.<method>

    def method(self, *args, **kwargs):
        # TODO switch to __make_request
        return self._make_request(method_name, args, kwargs)

    return method



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import logging
from collections import deque
from itertools import chain
from random import random

from tornado import gen

from ..schemes import DEFAULT as DEFAULT_SCHEME
from ..retry import (
    DEFAULT as DEFAULT_RETRY, DEFAULT_RETRY_LIMIT
)
from tchannel.event import EventType
from tchannel.glossary import DEFAULT_TIMEOUT
from ..context import get_current_context
from ..errors import NoAvailablePeerError
from ..errors import TChannelError
from ..zipkin.annotation import Endpoint
from ..zipkin.trace import Trace
from .connection import StreamConnection
from .request import Request
from .stream import InMemStream
from .stream import read_full
from .stream import maybe_stream
from .timeout import timeout

try:
    # included in Tornado 4.2
    from tornado.locks import Condition
except ImportError:  # pragma: no cover
    from toro import Condition

log = logging.getLogger('tchannel')


class PeerGroup(object):
    """A PeerGroup represents a collection of Peers.

    Requests routed through a PeerGroup can be sent to either a specific peer
    or a peer chosen at random.
    """

    def __init__(self, tchannel, score_threshold=None):
        """Initializes a new PeerGroup.

        :param tchannel:
            TChannel used for communication by this PeerGroup
        :param score_threshold:
            A value in the ``[0, 1]`` range. If specifiede, this requires that
            chosen peers havea score higher than this value when performing
            requests.
        """
        self.tchannel = tchannel

        self._score_threshold = score_threshold

        # Dictionary from hostport to Peer.
        self._peers = {}

        # Notified when a reset is performed. This allows multiple coroutines
        # to block on the same reset.
        self._resetting = False

        # We'll create a Condition here later. We want to avoid it right now
        # because it has a side-effect of scheduling some dummy work on the
        # ioloop, which prevents us from forking (if you're into that).
        self._reset_condition = None

    def __str__(self):
        return "<PeerGroup peers=%s>" % str(self._peers)

    @gen.coroutine
    def clear(self):
        """Reset this PeerGroup.

        This closes all connections to all known peers and forgets about
        these peers.

        :returns:
            A Future that resolves with a value of None when the operation
            has finished
        """
        if self._resetting:
            # If someone else is already resetting the PeerGroup, just block
            # on them to be finished.
            yield self._reset_condition.wait()
            raise gen.Return(None)

        self._resetting = True
        if self._reset_condition is None:
            self._reset_condition = Condition()

        try:
            for peer in self._peers.values():
                peer.close()
        finally:
            self._peers = {}
            self._resetting = False
            self._reset_condition.notify_all()

    def get(self, hostport):
        """Get a Peer for the given destination.

        A new Peer is added and returned if one does not already exist for the
        given host-port. Otherwise, the existing Peer is returned.
        """
        assert hostport, "hostport is required"
        if hostport not in self._peers:
            self._peers[hostport] = Peer(self.tchannel, hostport)
        return self._peers[hostport]

    def lookup(self, hostport):
        """Look up a Peer for the given host and port.

        Returns None if a Peer for the given host-port does not exist.
        """
        assert hostport, "hostport is required"
        return self._peers.get(hostport, None)

    def remove(self, hostport):
        """Delete the Peer for the given host port.

        Does nothing if a matching Peer does not exist.

        :returns: The removed Peer
        """
        assert hostport, "hostport is required"
        return self._peers.pop(hostport, None)

    def add(self, peer):
        """Add an existing Peer to this group.

        A peer for the given host-port must not already exist in the group.
        """
        assert peer, "peer is required"

        if isinstance(peer, basestring):
            # Assume strings are host-ports
            peer = Peer(self.tchannel, peer)

        assert peer.hostport not in self._peers, (
            "%s already has a peer" % peer.hostport
        )

        self._peers[peer.hostport] = peer

    @property
    def hosts(self):
        """Get all host-ports managed by this PeerGroup."""
        return self._peers.keys()

    @property
    def peers(self):
        """Get all Peers managed by this PeerGroup."""
        return self._peers.values()

    def request(self, service, hostport=None, **kwargs):
        """Initiate a new request through this PeerGroup.

        :param hostport:
            If specified, requests will be sent to the specific host.
            Otherwise, a known peer will be picked at random.
        :param service:
            Name of the service being called. Defaults to an empty string.
        :param service_threshold:
            If ``hostport`` was not specified, this specifies the score
            threshold at or below which peers will be ignored.
        """
        return PeerClientOperation(
            peer_group=self,
            service=service,
            hostport=hostport,
            **kwargs)

    def choose(self, hostport=None, score_threshold=None, blacklist=None):
        """Choose a Peer that matches the given criteria.

        The Peer with the highest score will be chosen.

        :param hostport:
            Specifies that the returned Peer must be for the given host-port.
            Without this, all peers managed by this PeerGroup are
            candidates. If this is present, ``score_threshold`` is ignored.
        :param score_threshold:
            If specified, Peers with a score equal to or below this will be
            ignored. Defaults to the value specified when the PeerGroup was
            initialized.
        :param blacklist:
            Peers on the blacklist won't be chosen.
        :returns:
            A Peer that matches all the requested criteria or None if no such
            Peer was found.
        """

        blacklist = blacklist or set()
        if hostport:
            return self.get(hostport)

        score_threshold = score_threshold or self._score_threshold or 0
        chosen_peer = None
        chosen_score = 0
        hosts = self._peers.viewkeys() - blacklist

        for host in hosts:
            peer = self.get(host)
            score = peer.state.score()

            if score <= score_threshold:
                continue

            if score > chosen_score:
                chosen_peer = peer
                chosen_score = score

        return chosen_peer


class Peer(object):
    """A Peer manages connections to or from a specific host-port."""

    __slots__ = (
        'tchannel',
        'state',
        'host',
        'port',

        '_out_conns',
        '_in_conns',
        '_connecting',
    )

    # Class used to create new outgoing connections.
    #
    # It must support a .outgoing method.
    connection_class = StreamConnection

    def __init__(self, tchannel, hostport, state=None):
        """Initialize a Peer

        :param tchannel:
            TChannel through which requests will be made.
        :param hostport:
            Host-port this Peer is for.
        :param state:
            State of the Peer. If given, this must be an instance of PeerState.
        """
        state = state or PeerHealthyState(self)

        assert hostport, "hostport is required"
        assert isinstance(state, PeerState), "state must be a PeerState"

        self.tchannel = tchannel
        self.state = state
        self.host, port = hostport.rsplit(':', 1)
        self.port = int(port)

        self._out_conns = deque()
        self._in_conns = deque()

        # This contains a future to the TornadoConnection if we're already in
        # the process of making an outgoing connection to the peer. This
        # helps avoid making multiple outgoing connections.
        self._connecting = None

    def connect(self):
        """Get a connection to this peer.

        If an connection to the peer already exists (either incoming or
        outgoing), that's returned. Otherwise, a new outgoing connection to
        this peer is created.

        :return:
            A future containing a connection to this host.
        """
        # Prefer recently created outgoing connections over everything else.
        conns = (
            conn for conn in chain(self._out_conns, self._in_conns)
            if not conn.closed
        )
        try:
            conn = next(conns)
        except StopIteration:
            pass
        else:
            # Wrap the connection in a Future
            return gen.maybe_future(conn)

        if self._connecting:
            # If we're in the process of connecting to the peer, just wait
            # and re-use that connection.
            return self._connecting

        conn_future = self._connecting = self.connection_class.outgoing(
            hostport=self.hostport,
            process_name=self.tchannel.process_name,
            serve_hostport=self.tchannel.hostport,
            handler=self.tchannel.receive_call,
            tchannel=self.tchannel,
        )

        def on_connect(_):
            if not conn_future.exception():
                # We don't actually need to handle the exception. That's on
                # the caller.

                connection = conn_future.result()
                self._out_conns.appendleft(connection)
            self._connecting = None

        conn_future.add_done_callback(on_connect)
        return conn_future

    def register_incoming(self, conn):
        assert conn, "conn is required"
        self._in_conns.append(conn)

        # TODO on-close cleanup

    @property
    def hostport(self):
        """The host-port this Peer is for."""
        return "%s:%d" % (self.host, self.port)

    @property
    def connections(self):
        """Returns a list of all connections for this peer.

        Incoming connections are listed first."""
        return list(chain(self._in_conns, self._out_conns))

    @property
    def outgoing_connections(self):
        """Returns a list of all outgoing connections for this peer."""
        return list(self._out_conns)

    @property
    def incoming_connections(self):
        """Returns a list of all incoming connections for this peer."""
        return list(self._in_conns)

    @property
    def is_ephemeral(self):
        """Whether this Peer is ephemeral."""
        return self.host == '0.0.0.0' and self.port == 0

    @property
    def connected(self):
        """Return True if this Peer is connected."""
        for conn in self.connections:
            if not conn.closed:
                return True
        return False

    def close(self):
        # TODO: Debounce like PeerGroup?
        for connection in self.connections:
            connection.close()


class PeerState(object):
    """Represents the state of the Peer."""

    __slots__ = ()

    def score(self):
        """Calculate the score of the Peer in this state."""
        return 0


class PeerHealthyState(PeerState):
    """Indicates that the Peer is in a healthy state.

    The score of healthy peers is a random value between 0.2 and 1.0. This
    allows random selection between multiple matches.
    """

    __slots__ = ('peer',)

    def __init__(self, peer):
        self.peer = peer

    def score(self):
        # Connected peers have a score in the range [0.2, 1.0) and all other
        # peers have a score in the range [0.0, 0.2). So, we will always
        # prefer peers that are already connected over peers that require new
        # connections.
        if self.peer.connected:
            return 0.2 + random() * 0.8
            # TODO this can be split between incoming and outgoing connections.
            # We probably want to prefer outgoing connections over incoming
            # connections.
        else:
            return 0.1 + random() * 0.1

            # TODO: It may be reasonable to allow the Peer or TChannel to
            # control randomness.


class PeerClientOperation(object):
    """Encapsulates client operations that can be performed against a peer."""

    def __init__(self,
                 peer_group,
                 service,
                 arg_scheme=None,
                 retry=None,
                 parent_tracing=None,
                 hostport=None,
                 score_threshold=None):
        """Initialize a new PeerClientOperation.

        :param peer_group:
            instance of PeerGroup that maintain all peers.
        :param service:
            Name of the service being called through this peer. Defaults to
            an empty string.
        :param arg_scheme
            arg scheme type
        :param retry
            retry type
        :param parent_tracing
            tracing span from parent request
        :param hostport
            remote server's host port.
        """
        assert peer_group, "peer group must not be None"
        service = service or ''

        self.peer_group = peer_group
        self.tchannel = peer_group.tchannel
        self.service = service
        self.parent_tracing = parent_tracing
        if not self.parent_tracing and get_current_context():
            self.parent_tracing = get_current_context().parent_tracing

        # TODO the term headers are reserved for application headers,
        # these are transport headers,
        self.headers = {
            'as': arg_scheme or DEFAULT_SCHEME,
            're': retry or DEFAULT_RETRY,
            'cn': self.tchannel.name,
        }

        # keep all arguments for retry purpose.
        # not retry if hostport is set.
        self._hostport = hostport
        self._score_threshold = score_threshold
        # service name is not stored in peer because the same peer may be
        # used to call multiple services if it's being used for request
        # forwarding

    def _choose(self, blacklist=None):
        peer = self.peer_group.choose(
            hostport=self._hostport,
            score_threshold=self._score_threshold,
            blacklist=blacklist,
        )

        return peer

    @gen.coroutine
    def send(
        self, arg1, arg2, arg3,
        headers=None,
        traceflag=None,
        retry_limit=None,
        ttl=None,
    ):
        """Make a request to the Peer.

        :param arg1:
            String or Stream containing the contents of arg1. If None, an empty
            stream is used.
        :param arg2:
            String or Stream containing the contents of arg2. If None, an empty
            stream is used.
        :param arg3:
            String or Stream containing the contents of arg3. If None, an empty
            stream is used.
        :param headers:
            Headers will be put int he message as protocol header.
        :param traceflag:
            Flag is for tracing.
        :param retry_limit:
           Maximum number of retries will perform on the message. If the number
           is 0, it means no retry.
        :param ttl:
            Timeout for each request (ms).
        :return:
            Future that contains the response from the peer.
        """

        # find a peer connection
        # If we can't find available peer at the first time, we throw
        # NoAvailablePeerError. Later during retry, if we can't find available
        # peer, we throw exceptions from retry not NoAvailablePeerError.
        peer = self._choose()
        if not peer:
            raise NoAvailablePeerError(
                "Can't find an available peer for '%s'" % self.service
            )

        connection = yield peer.connect()

        arg1, arg2, arg3 = (
            maybe_stream(arg1), maybe_stream(arg2), maybe_stream(arg3)
        )

        if retry_limit is None:
            retry_limit = DEFAULT_RETRY_LIMIT

        ttl = ttl or DEFAULT_TIMEOUT
        # hack to get endpoint from arg_1 for trace name
        arg1.close()
        endpoint = yield read_full(arg1)

        # TODO after adding stackcontext, get ride of this.
        if self.parent_tracing:
            parent_span_id = self.parent_tracing.span_id
            trace_id = self.parent_tracing.trace_id
        else:
            parent_span_id = None
            trace_id = None

        if traceflag is None:
            traceflag = self.tchannel.trace

        traceflag = traceflag() if callable(traceflag) else traceflag

        # set default transport headers
        headers = headers or {}
        for k, v in self.headers.iteritems():
            headers.setdefault(k, v)

        request = Request(
            service=self.service,
            argstreams=[InMemStream(endpoint), arg2, arg3],
            id=connection.next_message_id(),
            headers=headers,
            endpoint=endpoint,
            ttl=ttl,
            tracing=Trace(
                name=endpoint,
                trace_id=trace_id,
                parent_span_id=parent_span_id,
                endpoint=Endpoint(peer.host, peer.port, self.service),
                traceflags=traceflag,
            )
        )

        # only retry on non-stream request
        if request.is_streaming_request or self._hostport:
            retry_limit = 0

        if request.is_streaming_request:
            request.ttl = 0

        try:
            response = yield self.send_with_retry(
                request, peer, retry_limit, connection
            )
        except Exception as e:
            # event: on_exception
            self.tchannel.event_emitter.fire(
                EventType.on_exception, request, e,
            )
            raise

        log.debug("Got response %s", response)

        raise gen.Return(response)

    @gen.coroutine
    def _send(self, connection, req):
        # event: send_request
        self.tchannel.event_emitter.fire(EventType.before_send_request, req)
        response_future = connection.send_request(req)

        with timeout(response_future, req.ttl):
            try:
                response = yield response_future
            except TChannelError as error:
                # event: after_receive_error
                self.tchannel.event_emitter.fire(
                    EventType.after_receive_error, req, error,
                )
                raise
        # event: after_receive_response
        self.tchannel.event_emitter.fire(
            EventType.after_receive_response, req, response,
        )
        raise gen.Return(response)

    @gen.coroutine
    def send_with_retry(self, request, peer, retry_limit, connection):
        # black list to record all used peers, so they aren't chosen again.
        blacklist = set()
        for num_of_attempt in range(retry_limit + 1):
            try:
                response = yield self._send(connection, request)
                raise gen.Return(response)
            # Why are we retying on all errors????
            except TChannelError as error:
                blacklist.add(peer.hostport)
                (peer, connection) = yield self._prepare_for_retry(
                    request=request,
                    connection=connection,
                    protocol_error=error,
                    blacklist=blacklist,
                    num_of_attempt=num_of_attempt,
                    max_retry_limit=retry_limit,
                )

                if not connection:
                    raise error

    @gen.coroutine
    def _prepare_for_retry(
        self,
        request,
        connection,
        protocol_error,
        blacklist,
        num_of_attempt,
        max_retry_limit,
    ):

        self.clean_up_outgoing_request(request, connection, protocol_error)
        if not self.should_retry(request, protocol_error,
                                 num_of_attempt, max_retry_limit):
            raise gen.Return((None, None))

        result = yield self.prepare_next_request(request, blacklist)
        raise gen.Return(result)

    @gen.coroutine
    def prepare_next_request(self, request, blacklist):
        # find new peer
        peer = self._choose(blacklist=blacklist,)

        # no peer is available
        if not peer:
            raise gen.Return((None, None))

        connection = yield peer.connect()
        # roll back request
        request.rewind(connection.next_message_id())

        raise gen.Return((peer, connection))

    @staticmethod
    def should_retry(request, error, num_of_attempt, max_retry_limit):
        return (
            request.should_retry_on_error(error) and
            num_of_attempt != max_retry_limit
        )

    @staticmethod
    def clean_up_outgoing_request(request, connection, error):
        # stop the outgoing request
        request.set_exception(error)
        # remove from pending request list
        connection.remove_outstanding_request(request)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, print_function, unicode_literals

import sys
import types
from functools import partial

import thriftrw
from tornado import gen
from tornado.util import raise_exc_info

from tchannel.status import OK, FAILED
from tchannel.errors import OneWayNotSupportedError
from tchannel.errors import ValueExpectedError
from tchannel.response import Response, response_from_mixed
from tchannel.serializer.thrift import ThriftRWSerializer

from .module import ThriftRequest


def load(path, service=None, hostport=None, module_name=None):
    """Loads the Thrift file at the specified path.

    .. note::

        This functionality is experimental and subject to change. We expect to
        mark it as stable in a future version.

    The file is compiled in-memory and a Python module containing the result
    is returned. It may be used with ``TChannel.thrift``. For example,

    .. code-block:: python

        from tchannel import TChannel, thrift

        # Load our server's interface definition.
        donuts = thrift.load(path='donuts.thrift')

        # We need to specify a service name or hostport because this is a
        # downstream service we'll be calling.
        coffee = thrift.load(path='coffee.thrift', service='coffee')

        tchannel = TChannel('donuts')

        @tchannel.thrift.register(donuts.DonutsService)
        @tornado.gen.coroutine
        def submitOrder(request):
            args = request.body

            if args.coffee:
                yield tchannel.thrift(
                    coffee.CoffeeService.order(args.coffee)
                )

            # ...

    The returned module contains, one top-level type for each struct, enum,
    union, exeption, and service defined in the Thrift file. For each service,
    the corresponding class contains a classmethod for each function defined
    in that service that accepts the arguments for that function and returns a
    ``ThriftRequest`` capable of being sent via ``TChannel.thrift``.

    Note that the ``path`` accepted by ``load`` must be either an absolute
    path or a path relative to the *the current directory*. If you need to
    refer to Thrift files relative to the Python module in which ``load`` was
    called, use the ``__file__`` magic variable.

    .. code-block:: python

        # Given,
        #
        #   foo/
        #     myservice.thrift
        #     bar/
        #       x.py
        #
        # Inside foo/bar/x.py,

        path = os.path.join(
            os.path.dirname(__file__), '../myservice.thrift'
        )

    The returned value is a valid Python module. You can install the module by
    adding it to the ``sys.modules`` dictionary. This will allow importing
    items from this module directly. You can use the ``__name__`` magic
    variable to make the generated module a submodule of the current module.
    For example,

    .. code-block:: python

        # foo/bar.py

        import sys
        from tchannel import thrift

        donuts = = thrift.load('donuts.thrift')
        sys.modules[__name__ + '.donuts'] = donuts

    This installs the module generated for ``donuts.thrift`` as the module
    ``foo.bar.donuts``. Callers can then import items from that module
    directly. For example,

    .. code-block:: python

        # foo/baz.py

        from foo.bar.donuts import DonutsService, Order

        def baz(tchannel):
            return tchannel.thrift(
                DonutsService.submitOrder(Order(..))
            )

    :param str service:
        Name of the service that the Thrift file represents. This name will be
        used to route requests through Hyperbahn.
    :param str path:
        Path to the Thrift file. If this is a relative path, it must be
        relative to the current directory.
    :param str hostport:
        Clients can use this to specify the hostport at which the service can
        be found. If omitted, TChannel will route the requests through known
        peers. This value is ignored by servers.
    :param str module_name:
        Name used for the generated Python module. Defaults to the name of the
        Thrift file.
    """
    # TODO replace with more specific exceptions
    # assert service, 'service is required'
    # assert path, 'path is required'

    # Backwards compatibility for callers passing in service name as first arg.
    if not path.endswith('.thrift'):
        service, path = path, service

    module = thriftrw.load(path=path, name=module_name)
    return TChannelThriftModule(service, module, hostport)


class TChannelThriftModule(types.ModuleType):
    """Wraps the ``thriftrw``-generated module.

    Wraps service classes with ``Service`` and exposes everything else from
    the module as-is.
    """

    def __init__(self, service, module, hostport=None):
        """Initialize a TChannelThriftModule.

        :param str service:
            Name of the service this module represents. This name will be used
            for routing over Hyperbahn.
        :param module:
            Module generated by ``thriftrw`` for a Thrift file.
        :param str hostport:
            This may be specified if the caller is a client and wants all
            requests sent to a specific address.
        """

        self.service = service
        self.module = module
        self.hostport = hostport

        for service_cls in self.module.services:
            name = service_cls.service_spec.name
            setattr(self, name, Service(service_cls, self))

    def __getattr__(self, name):
        return getattr(self.module, name)

    def __str__(self):
        return 'TChannelThriftModule(%s, %s)' % (self.service, self.module)

    __repr__ = __str__


class Service(object):
    """Wraps service classes generated by thriftrw.

    Exposes all functions of the service.
    """

    def __init__(self, cls, module):
        self._module = module
        self._cls = cls
        self._spec = cls.service_spec

        for func_spec in self._spec.functions:
            setattr(self, func_spec.name, Function(func_spec, self))

    @property
    def name(self):
        """Name of the Thrift service this object represents."""
        return self._spec.name

    def __str__(self):
        return 'Service(%s)' % self.name

    __repr__ = __str__


class Function(object):
    """Wraps a ServiceFunction generated by thriftrw.

    Acts as a callable that will construct ThriftRequests.
    """

    __slots__ = (
        'spec', 'func', 'service', 'request_cls', 'response_cls'
    )

    def __init__(self, func_spec, service):
        self.spec = func_spec
        self.func = func_spec.surface
        self.service = service

        self.request_cls = self.func.request
        self.response_cls = self.func.response

    @property
    def endpoint(self):
        """Endpoint name for this function."""
        return '%s::%s' % (self.service.name, self.func.name)

    @property
    def oneway(self):
        """Whether this function is oneway."""
        return self.spec.oneway

    def __call__(self, *args, **kwargs):
        if self.oneway:
            raise OneWayNotSupportedError(
                'TChannel+Thrift does not currently support oneway '
                'procedures.'
            )

        if not (
            self.service._module.hostport or
            self.service._module.service
        ):
            raise ValueError(
                "No 'service' or 'hostport' provided to " +
                str(self)
            )

        module = self.service._module
        call_args = self.request_cls(*args, **kwargs)

        return ThriftRWRequest(
            module=module,
            service=module.service,
            endpoint=self.endpoint,
            result_type=self.response_cls,
            call_args=call_args,
            hostport=module.hostport,
        )

    def __str__(self):
        return 'Function(%s)' % self.endpoint

    __repr__ = __str__


def register(dispatcher, service, handler=None, method=None):
    """
    :param dispatcher:
        RequestDispatcher against which the new endpoint will be registered.
    :param Service service:
        Service object representing the service whose endpoint is being
        registered.
    :param handler:
        A function implementing the given Thrift function.
    :param method:
        If specified, name of the method being registered. Defaults to the
        name of the ``handler`` function.
    """

    def decorator(method, handler):
        if not method:
            method = handler.__name__

        function = getattr(service, method, None)
        assert function, (
            'Service "%s" does not define method "%s"' % (service.name, method)
        )
        assert not function.oneway

        handler = build_handler(function, handler)
        dispatcher.register(
            function.endpoint,
            handler,
            ThriftRWSerializer(service._module, function.request_cls),
            ThriftRWSerializer(service._module, function.response_cls),
        )
        return handler

    if handler is None:
        return partial(decorator, method)
    else:
        return decorator(method, handler)


def build_handler(function, handler):
    # response_cls is a class that represents the response union for this
    # function. It accepts one parameter for each exception defined on the
    # method and another parameter 'success' for the result of the call. The
    # success kwarg is absent if the function doesn't return anything.
    response_cls = function.response_cls
    response_spec = response_cls.type_spec

    @gen.coroutine
    def handle(request):
        # kwargs for this function's response_cls constructor
        response_kwargs = {}
        status = OK

        try:
            response = yield gen.maybe_future(handler(request))
        except Exception as e:
            response = Response()

            for exc_spec in response_spec.exception_specs:
                # Each exc_spec is a thriftrw.spec.FieldSpec. The spec
                # attribute on that is the TypeSpec for the Exception class
                # and the surface on the TypeSpec is the exception class.
                exc_cls = exc_spec.spec.surface
                if isinstance(e, exc_cls):
                    status = FAILED
                    response_kwargs[exc_spec.name] = e
                    break
            else:
                raise_exc_info(sys.exc_info())
        else:
            response = response_from_mixed(response)

            if response_spec.return_spec is not None:
                assert response.body is not None, (
                    'Expected a value to be returned for %s, '
                    'but recieved None - only void procedures can '
                    'return None.' % function.endpoint
                )
                response_kwargs['success'] = response.body

        response.status = status
        response.body = response_cls(**response_kwargs)
        raise gen.Return(response)

    handle.__name__ = function.spec.name

    return handle


class ThriftRWRequest(ThriftRequest):

    def __init__(self, module, **kwargs):
        kwargs['serializer'] = ThriftRWSerializer(
            module, kwargs['result_type']
        )
        super(ThriftRWRequest, self).__init__(**kwargs)

    def read_body(self, body):
        response_spec = self.result_type.type_spec

        for exc_spec in response_spec.exception_specs:
            exc = getattr(body, exc_spec.name)
            if exc is not None:
                raise exc

        # success - non-void
        if response_spec.return_spec is not None:
            if body.success is None:
                raise ValueExpectedError(
                    'Expected a value to be returned for %s, '
                    'but recieved None - only void procedures can '
                    'return None.' % self.endpoint
                )

            return body.success

        # success - void
        else:
            return None



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from .exceptions import VCRError

__all__ = ['RecordMode']


class _RecordMode(object):
    """A _RecordMode dictates a cassette's behavior on recording and replays.

    :param name:
        Name of the record mode.
    :param replayable:
        Whether this record mode allows requests to be replayed.
    :param can_record:
        Whether this record mode allows new interactions to be recorded. This
        may be a boolean or a function that accepts the cassette and returns a
        boolean.
    :param save_unplayed:
        Whether the cassette remembers previously saved unplayed interactions
        when the cassette is saved again. This is useful when the record mode
        needs to forget unused interactions.
    """

    __slots__ = ('name', 'replayable', 'can_record', 'save_unplayed')

    def __init__(self, name, replayable, can_record, save_unplayed):
        self.name = name
        self.replayable = replayable
        self.save_unplayed = save_unplayed
        if not callable(can_record):
            self.can_record = (lambda _: can_record)
        else:
            self.can_record = can_record


class RecordMode(object):
    """
    Record modes dictate how a cassette behaves when interactions are replayed
    or recorded. The following record modes are supported.

    .. autoattribute:: ONCE
       :annotation: = 'once'

    .. autoattribute:: NEW_EPISODES
       :annotation: = 'new_episodes'

    .. autoattribute:: NONE
       :annotation: = 'none'

    .. autoattribute:: ALL
       :annotation: = 'all'
    """

    #: If the YAML file did not exist, record new interactions and save them.
    #: If the YAML file already existed, replay existing interactions but
    #: disallow any new interactions. This is the default and usually what you
    #: want.
    ONCE = _RecordMode(
        name='once',
        replayable=True,
        can_record=(lambda c: not c.existed),
        save_unplayed=True,
    )

    #: Replay existing interactions and allow recording new ones. This is
    #: usually undesirable since it reduces predictability in tests.
    NEW_EPISODES = _RecordMode(
        name='new_episodes',
        replayable=True,
        can_record=True,
        save_unplayed=True,
    )

    #: Replay existing interactions and disallow any new interactions.  This
    #: is a good choice for tests whose behavior is unlikely to change in the
    #: near future. It ensures that those tests don't accidentally start
    #: making new requests.
    NONE = _RecordMode(
        name='none',
        replayable=True,
        can_record=False,
        save_unplayed=True,
    )

    #: Do not replay anything and record all new interactions. Forget all
    #: existing interactions. This may be used to record everything anew.
    ALL = _RecordMode(
        name='all',
        replayable=False,
        can_record=True,
        save_unplayed=False,
    )

    @classmethod
    def from_name(cls, name):
        """Get a RecordMode by its name."""
        name = name.lower()

        if name not in _MODES:
            raise VCRError(
                'Invalid record mode %s. It must be one of "once", "none", '
                '"all", or "new_episodes". Check the documentation for more '
                'information' % repr(name)
            )
        return _MODES[name]


_MODES = {m.name.lower(): m for m in [
    RecordMode.ONCE, RecordMode.NEW_EPISODES, RecordMode.NONE, RecordMode.ALL
]}



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#
# Autogenerated by Thrift Compiler (0.9.2)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,slots,dynamic
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
from ttypes import *
from thrift.Thrift import TProcessor
from thrift.protocol.TBase import TBase, TExceptionBase


class Iface(object):
  def submit(self, span):
    """
    Parameters:
     - span
    """
    pass


class Client(Iface):
  def __init__(self, iprot, oprot=None):
    self._iprot = self._oprot = iprot
    if oprot is not None:
      self._oprot = oprot
    self._seqid = 0

  def submit(self, span):
    """
    Parameters:
     - span
    """
    self.send_submit(span)
    return self.recv_submit()

  def send_submit(self, span):
    self._oprot.writeMessageBegin('submit', TMessageType.CALL, self._seqid)
    args = submit_args()
    args.span = span
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_submit(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = submit_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "submit failed: unknown result");


class Processor(Iface, TProcessor):
  def __init__(self, handler):
    self._handler = handler
    self._processMap = {}
    self._processMap["submit"] = Processor.process_submit

  def process(self, iprot, oprot):
    (name, type, seqid) = iprot.readMessageBegin()
    if name not in self._processMap:
      iprot.skip(TType.STRUCT)
      iprot.readMessageEnd()
      x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
      oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
      x.write(oprot)
      oprot.writeMessageEnd()
      oprot.trans.flush()
      return
    else:
      self._processMap[name](self, seqid, iprot, oprot)
    return True

  def process_submit(self, seqid, iprot, oprot):
    args = submit_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = submit_result()
    result.success = self._handler.submit(args.span)
    oprot.writeMessageBegin("submit", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES

class submit_args(TBase):
  """
  Attributes:
   - span
  """

  __slots__ = [ 
    'span',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'span', (Span, Span.thrift_spec), None, ), # 1
  )

  def __init__(self, span=None,):
    self.span = span

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.span)
    return value


class submit_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.STRUCT, 'success', (Response, Response.thrift_spec), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value




# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import json
import logging

from threading import Lock
from tornado import gen

from . import schemes
from . import transport
from . import retry
from .context import get_current_context
from .errors import AlreadyListeningError
from .glossary import DEFAULT_TIMEOUT
from .health import health
from .health import Meta
from .response import Response, TransportHeaders
from .tornado import TChannel as DeprecatedTChannel
from .tornado.dispatch import RequestDispatcher as DeprecatedDispatcher

log = logging.getLogger('tchannel')

__all__ = ['TChannel']


class TChannel(object):
    """Manages connections and requests to other TChannel services.

    Usage for a JSON client/server:

    .. code:: python

        tchannel = TChannel(name='foo')

        @tchannel.json.register
        def handler(request):
            return {'foo': 'bar'}

        response = yield tchannel.json(
            service='some-service',
            endpoint='endpoint',
            headers={'req': 'headers'},
            body={'req': 'body'},
        )

    :cvar thrift:
        Make Thrift requests over TChannel and register Thrift handlers.
    :vartype thrift: ThriftArgScheme

    :cvar json:
        Make JSON requests over TChannel and register JSON handlers.
    :vartype json: JsonArgScheme

    :cvar raw:
        Make requests and register handles that pass raw bytes.
    :vartype raw: RawArgScheme

    """

    def __init__(self, name, hostport=None, process_name=None,
                 known_peers=None, trace=False):
        """
        **Note:** In general only one ``TChannel`` instance should be used at a
        time. Multiple ``TChannel`` instances are not advisable and could
        result in undefined behavior.

        :param string name:
            How this application identifies itself. This is the name callers
            will use to make contact, it is also what your downstream services
            will see in their metrics.

        :param string hostport:
            An optional host/port to serve on, e.g., ``"127.0.0.1:5555``. If
            not provided an ephemeral port will be used. When advertising on
            Hyperbahn you callers do not need to know your port.
        """

        # until we move everything here,
        # lets compose the old tchannel
        self._dep_tchannel = DeprecatedTChannel(
            name=name,
            hostport=hostport,
            process_name=process_name,
            known_peers=known_peers,
            trace=trace,
            dispatcher=DeprecatedDispatcher(_handler_returns_response=True),
        )

        self.name = name

        # set arg schemes
        self.raw = schemes.RawArgScheme(self)
        self.json = schemes.JsonArgScheme(self)
        self.thrift = schemes.ThriftArgScheme(self)
        self._listen_lock = Lock()
        # register default health endpoint
        self.thrift.register(Meta)(health)

    def is_listening(self):
        return self._dep_tchannel.is_listening()

    @property
    def hooks(self):
        return self._dep_tchannel.hooks

    @gen.coroutine
    def call(
        self,
        scheme,
        service,
        arg1,
        arg2=None,
        arg3=None,
        timeout=None,
        retry_on=None,
        retry_limit=None,
        hostport=None,
        shard_key=None,
        trace=None,
    ):
        """Make low-level requests to TChannel services.

        **Note:** Usually you would interact with a higher-level arg scheme
        like :py:class:`tchannel.schemes.JsonArgScheme` or
        :py:class:`tchannel.schemes.ThriftArgScheme`.
        """

        # TODO - dont use asserts for public API
        assert format, "format is required"
        assert service, "service is required"
        assert arg1, "arg1 is required"

        # default args
        if arg2 is None:
            arg2 = ""
        if arg3 is None:
            arg3 = ""
        if timeout is None:
            timeout = DEFAULT_TIMEOUT
        if retry_on is None:
            retry_on = retry.DEFAULT
        if retry_limit is None:
            retry_limit = retry.DEFAULT_RETRY_LIMIT

        # TODO - allow filters/steps for serialization, tracing, etc...
        context = get_current_context()

        # calls tchannel.tornado.peer.PeerClientOperation.__init__
        operation = self._dep_tchannel.request(
            service=service,
            hostport=hostport,
            arg_scheme=scheme,
            retry=retry_on,
            parent_tracing=context.parent_tracing if context else None
        )

        # fire operation
        transport_headers = {
            transport.SCHEME: scheme,
            transport.CALLER_NAME: self.name,
        }
        if shard_key:
            transport_headers[transport.SHARD_KEY] = shard_key

        # If we got some parent tracing info we always want to propagate it
        # along. Otherwise use the ``trace`` parameter that was passed in. If
        # **that** wasn't provided, fall back to the TChannel default.
        trace_override = self._dep_tchannel.trace if trace is None else trace
        traceflag = bool(context.parent_tracing if context else trace_override)

        response = yield operation.send(
            arg1=arg1,
            arg2=arg2,
            arg3=arg3,
            headers=transport_headers,
            retry_limit=retry_limit,
            ttl=timeout,
            traceflag=traceflag,
        )

        # unwrap response
        body = yield response.get_body()
        headers = yield response.get_header()
        t = transport.to_kwargs(response.headers)
        t = TransportHeaders(**t)
        result = Response(
            body=body,
            headers=headers,
            transport=t
        )

        raise gen.Return(result)

    def listen(self, port=None):
        with self._listen_lock:
            if self._dep_tchannel.is_listening():
                listening_port = int(self.hostport.rsplit(":")[1])
                if port and port != listening_port:
                    raise AlreadyListeningError(
                        "TChannel server is already listening on port: %d"
                        % listening_port
                    )
                else:
                    return
            return self._dep_tchannel.listen(port)

    @property
    def hostport(self):
        return self._dep_tchannel.hostport

    def register(self, scheme, endpoint=None, handler=None, **kwargs):

        def decorator(fn):

            # assert handler is None, "can't handler when using as decorator"

            if endpoint is None:
                e = fn.__name__
            else:
                e = endpoint

            return self._dep_tchannel.register(
                endpoint=e,
                scheme=scheme,
                handler=fn,
                **kwargs
            )

        if handler is None:
            return decorator
        else:
            return decorator(handler)

    @gen.coroutine
    def advertise(self, routers=None, name=None, timeout=None,
                  router_file=None):
        """Advertise with Hyperbahn.

        After a successful advertisement, Hyperbahn will establish long-lived
        connections with your application. These connections are used to load
        balance inbound and outbound requests to other applications on the
        Hyperbahn network.

        Re-advertisement happens periodically after calling this method (every
        minute). Hyperbahn will eject us from the network if it doesn't get a
        re-advertise from us after 5 minutes.

        :param list routers:
            A seed list of known Hyperbahn addresses to attempt contact with.
            Entries should be of the form ``"host:port"``.

        :param string name:
            The name your application identifies itself as. This is usually
            unneeded because in the common case it will match the ``name`` you
            initialized the ``TChannel`` instance with. This is the identifier
            other services will use to make contact with you.

        :param timeout:
            The timeout (in seconds) for the initial advertise attempt.
            Defaults to 30 seconds.

        :param router_file:
            The host file that contains the routers information. The file
            should contain a JSON stringified format of the routers parameter.
            Either routers or router_file should be provided. If both provided,
            a ValueError will be raised.

        :returns:
            A future that resolves to the remote server's response after the
            first advertise finishes.

        :raises TimeoutError:
            When unable to make our first advertise request to Hyperbahn.
            Subsequent requests may fail but will be ignored.
        """
        if routers is not None and router_file is not None:
            raise ValueError(
                'Only one of routers and router_file can be provided.')

        if routers is None and router_file is not None:
            # should just let the exceptions fly
            try:
                with open(router_file, 'r') as json_data:
                    routers = json.load(json_data)
            except (IOError, OSError, ValueError):
                log.exception('Failed to read seed routers list.')
                raise

        dep_result = yield self._dep_tchannel.advertise(
            routers=routers,
            name=name,
            timeout=timeout
        )

        body = yield dep_result.get_body()
        headers = yield dep_result.get_header()
        response = Response(json.loads(body), headers or {})

        raise gen.Return(response)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import argparse
import collections
import contextlib
import cProfile
import itertools
import logging
import pstats
import sys
import time

import tornado.ioloop

from .tornado import TChannel

log = logging.getLogger('tchannel')


def parse_args(args=None):
    args = args or sys.argv[1:]

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--host",
        dest="host",
        default="localhost:8888/",
        nargs='+',
        help="Hostname, port, and optional endpoint (arg1) to hit.",
    )

    parser.add_argument(
        "-d", "--body",
        dest="body",
        default=[None],
        nargs='*',
        help=(
            "arg3. Can be specified multiple times to trigger simultaneous "
            "requests."
        ),
    )

    parser.add_argument(
        "-s", "--service",
        dest="service",
        default=None,
        help=(
            "Name of destination service (for tchannel routing)."
        ),
    )

    parser.add_argument(
        "-H", "--headers",
        dest="headers",
        default=[None],
        nargs='*',
        help=(
            "arg2. Can be speecified multiple time to trigger simultaneous "
            "requests."
        ),
    )

    parser.add_argument(
        "-r", "--rps",
        dest="rps",
        type=int,
        default=None,
        help=(
            "Throttle outgoing requests to this many per second."
        ),
    )

    parser.add_argument(
        "-v", "--verbose",
        dest="verbose",
        action="store_true"
    )

    parser.add_argument(
        "-q", "--quiet",
        dest="quiet",
        action="store_true",
        help="Don't display request/response information.",
    )

    parser.add_argument(
        "--batch-size",
        dest="batch_size",
        type=int,
        default=100,
        help="Maximum number of outstanding requests.",
    )

    parser.add_argument(
        "--profile",
        dest="profile",
        action="store_true"
    )

    args = parser.parse_args(args)

    # Allow a body/header to specified once and shared across multiple
    # requests.
    if args.headers and args.body and len(args.headers) != len(args.body):
        if len(args.headers) == 1:
            args.headers = args.headers * len(args.body)

        elif len(args.body) == 1:
            args.body = args.body * len(args.headers)

        else:
            parser.error(
                "Multiple header/body arguments given "
                "but not of the same degree."
            )

    if len(args.host) != max(1, len(args.headers), len(args.body)):
        if len(args.host) != 1:
            parser.error(
                "Number of hosts specified doesn't agree with the number of "
                "header/body arguments."
            )

        args.host = args.host * max(len(args.headers), len(args.body))

    # Transform something like "localhost:8888" into "localhost:8888/" so we
    # consider it as the '' endpoint.
    args.host = (h if '/' in h else h + '/' for h in args.host)

    return args


def chunk(iterable, n):
    while True:
        chunk = tuple(itertools.islice(iterable, n))
        if not chunk:
            return
        yield chunk


@tornado.gen.coroutine
def multi_tcurl(
    tchannel,
    hostports,
    headers,
    bodies,
    service,
    profile=False,
    rps=None,
    quiet=False,
    batch_size=100,
):
    all_requests = getattr(itertools, 'izip', zip)(hostports, headers, bodies)

    with timing(profile):
        for requests in chunk(all_requests, batch_size):
            futures = []

            for hostport, header, body in requests:
                futures.append(
                    tcurl(tchannel, hostport, header, body, service, quiet)
                )

                if rps:
                    yield tornado.gen.sleep(1.0 / rps)

            wait_iterator = tornado.gen.WaitIterator(*futures)
            results = []

            while not wait_iterator.done():
                try:
                    results.append((yield wait_iterator.next()))
                except Exception:
                    raise

    raise tornado.gen.Return(results)


@tornado.gen.coroutine
def tcurl(tchannel, hostport, headers, body, service, quiet=False):
    host, endpoint = hostport.split('/', 1)

    if not quiet:
        log.debug("> Host: %s" % host)
        log.debug("> Arg1: %s" % endpoint)
        log.debug("> Arg2: %s" % headers)
        log.debug("> Arg3: %s" % body)

    request = tchannel.request(host, service)

    response = yield request.send(
        endpoint,
        headers,
        body,
    )

    if not quiet:
        log.debug("< Host: %s" % host)
        log.debug("<  Msg: %s" % response.id)

    raise tornado.gen.Return(response)


@contextlib.contextmanager
def timing(profile=False):
    start = time.time()

    profiler = None

    if profile:
        profiler = cProfile.Profile()
        profiler.enable()

    info = collections.Counter()

    yield info

    if profiler:
        profiler.disable()
        profiler.create_stats()
        stats = pstats.Stats(profiler)
        stats.sort_stats('time').print_stats(15)

    stop = time.time()

    log.info("took %.2fs seconds" % (stop - start))


@tornado.gen.coroutine
def main(argv=None):
    args = parse_args(argv)

    logging.basicConfig(
        format="%(name)s[%(process)s] %(levelname)s: %(message)s",
        stream=sys.stdout,
        level=logging.INFO,
    )

    log.setLevel(logging.INFO)
    if args.verbose:
        log.setLevel(logging.DEBUG)

    tchannel = TChannel(name='tcurl.py')

    results = yield multi_tcurl(
        tchannel,
        args.host,
        args.headers,
        args.body,
        args.service,
        profile=args.profile,
        rps=args.rps,
        quiet=args.quiet,
        batch_size=args.batch_size,
    )

    raise tornado.gen.Return(results)


def start_ioloop():  # pragma: no cover
    ioloop = tornado.ioloop.IOLoop.current()
    ioloop.run_sync(main)


if __name__ == '__main__':  # pragma: no cover
    start_ioloop()



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import


import inspect


def get_service_methods(iface):
    """Get a list of methods defined in the interface for a Thrift service.

    :param iface:
        The Thrift-generated Iface class defining the interface for the
        service.
    :returns:
        A set containing names of the methods defined for the service.
    """
    methods = inspect.getmembers(iface, predicate=inspect.ismethod)

    return set(
        name for (name, method) in methods if not name.startswith('__')
    )


def get_module_name(module):
    name = module.__name__.rsplit('.', 1)[-1]

    return name



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import collections
import math
import time

from tchannel.zipkin.thrift import constants

Endpoint = collections.namedtuple(
    'Endpoint',
    'ipv4, port, service_name'
)


_AnnotationBase = collections.namedtuple(
    'Annotation',
    'name, value, annotation_type, endpoint',
)


class Annotation(_AnnotationBase):

    def __new__(cls, name, value, annotation_type, endpoint=None):
        """
        :param name:
            Name of this annotation.
        :type name: str

        :param value:
            A value of the appropriate type based on ``annotation_type``.

        :param annotation_type:
            The expected type of our ``value``. Expected values are
            ``'string'`` and ``'bytes'``.
        :type annotation_type: str

        :param endpoint:
            An optional :py:class:`Endpoint` to associate with
            this annotation.
        :type endpoint: Endpoint or None
        """
        return super(Annotation, cls).__new__(
            cls,
            name,
            value,
            annotation_type,
            endpoint,
        )


def timestamp(name, ts=None):
    if ts is None:
        ts = math.trunc(time.time() * 1000)

    return Annotation(name, ts, 'timestamp')


def client_send(ts=None):
    return timestamp(constants.CLIENT_SEND, ts)


def client_recv(ts=None):
    return timestamp(constants.CLIENT_RECV, ts)


def server_send(ts=None):
    return timestamp(constants.SERVER_SEND, ts)


def server_recv(ts=None):
    return timestamp(constants.SERVER_RECV, ts)


def string(name, value):
    return Annotation(name, value, 'string')


def bytes(name, value):
    return Annotation(name, value, 'bytes')



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from thrift.Thrift import TType
from thrift.protocol.TBase import TBase

from .exceptions import VCRError


def to_native(ttype, data, extra=None):
    """Converts types into native classes and objects.

    :param ttype:
        Type of value to construct
    :param data:
        Raw Thrift value. For primitive types, this is usually unchanged. For
        structs it's a dictionary.
    :param extra:
        Extra information about the ttype.
    """
    if TType.LIST == ttype:
        return [to_native(extra[0], x, extra[1]) for x in data]
    elif TType.MAP == ttype:
        return {
            to_native(extra[0], k, extra[1]): to_native(extra[2], v, extra[3])
            for k, v in data.items()
        }
    elif TType.SET == ttype:
        return set([to_native(extra[0], x, extra[1]) for x in data])
    elif TType.STRUCT == ttype:
        cls = extra[0]
        spec = extra[1]
        kwargs = {}
        for entry in spec:
            if entry is None:
                continue
            name = entry[2]
            if name in data:
                kwargs[name] = to_native(entry[1], data[name], entry[3])
        return cls(**kwargs)
    else:
        return data


def to_primitive(ttype, value, extra=None):
    """Convert arbitrary Thrift values into serializable primitives.

    :param ttype:
        Type of the value being converted
    :param value:
        Value to convert
    :param extra:
        Extra information about the type
    """
    if TType.LIST == ttype:
        return [to_primitive(extra[0], x, extra[1]) for x in value]
    elif TType.MAP == ttype:
        return {
            to_primitive(extra[0], k, extra[1]):
                to_primitive(extra[2], v, extra[3])
            for k, v in value.items()
        }
    elif TType.SET == ttype:
        return [to_primitive(extra[0], x, extra[1]) for x in value]
    elif TType.STRUCT == ttype:
        spec = extra[1]
        data = {}
        for entry in spec:
            if entry is None:
                continue
            name = entry[2]
            attr = getattr(value, name)
            if attr is not None:
                data[name] = to_primitive(entry[1], attr, entry[3])
        return data
    else:
        return value


def enum_to_name(cls, value):
    return cls._VALUES_TO_NAMES[value]


def enum_values(cls):
    return cls._NAMES_TO_VALUES.values()


class VCRThriftMeta(type):

    def __new__(cls, name, bases, dct):
        if '_VALUES_TO_NAMES' in dct:
            # Provide a way to convert an enum value into its name.
            dct['to_name'] = classmethod(enum_to_name)
            dct['values'] = classmethod(enum_values)

        return type.__new__(cls, name, bases, dct)


class VCRThriftBase(TBase, object):
    """Base class for generated Thrift types.

    Adds the following:

    - ``to_primitive`` method to convert objects into serializable primitives.
    - ``to_native`` class method to construct objects of the class from
        primitives.
    - Readable ``__str__`` and ``__repr__`` methods.
    - For enums, a ``to_name`` class method that maps enum values to the enum
      value name, and a ``values`` method that returns a list of values
      defined for the enum.
    """

    __metaclass__ = VCRThriftMeta

    # TODO do we want to create copies of these types with manual mappers
    # instead?

    @classmethod
    def to_native(cls, data):
        return to_native(TType.STRUCT, data, (cls, cls.thrift_spec))

    def to_primitive(self):
        return to_primitive(
            TType.STRUCT, self, (self.__class__, self.thrift_spec)
        )

    # A more useful __str__ implementation

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self.to_primitive()))

    __repr__ = __str__


class VCRThriftError(VCRThriftBase, VCRError):
    "Base class for generated exceptions."""


__all__ = ['VCRThriftBase', 'VCRThriftError']



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import inspect
from collections import namedtuple

from thrift import Thrift
from tornado import gen

from tchannel import schemes
from tchannel.errors import OneWayNotSupportedError

from ..serializer.thrift import ThriftSerializer
from .reflection import get_service_methods

# Generated clients will use this base class.
_ClientBase = namedtuple(
    '_ClientBase',
    'tchannel hostport service trace protocol_headers'
)


def client_for(service, service_module, thrift_service_name=None):
    """Build a client class for the given Thrift service.

    The generated class accepts a TChannel and an optional hostport as
    initialization arguments.

    Given ``CommentService`` defined in ``comment.thrift`` and registered with
    Hyperbahn under the name "comment", here's how this may be used:

    .. code-block:: python

        from comment import CommentService

        CommentServiceClient = client_for("comment", CommentService)

        @gen.coroutine
        def post_comment(articleId, msg, hostport=None):
            client = CommentServiceClient(tchannel, hostport)
            yield client.postComment(articleId, CommentService.Comment(msg))

    :param service:
        Name of the Hyperbahn service being called. This is the name with
        which the service registered with Hyperbahn.
    :param service_module:
        The Thrift-generated module for that service. This usually has the
        same name as defined for the service in the IDL.
    :param thrift_service_name:
        If the Thrift service has a different name than its module, use this
        parameter to specify it.
    :returns:
        An object with the same interface as the service that uses the given
        TChannel to call the service.
    """
    assert service_module, 'service_module is required'
    service = service or ''  # may be blank for non-hyperbahn use cases
    if not thrift_service_name:
        thrift_service_name = service_module.__name__.rsplit('.', 1)[-1]

    method_names = get_service_methods(service_module.Iface)

    def new(cls, tchannel, hostport=None, trace=False, protocol_headers=None):
        """
        :param tchannel:
            TChannel through which the requests will be sent.
        :param hostport:
            Address of the machine to which the requests will be sent, or None
            if the TChannel will do peer selection on a per-request basis.
        :param trace:
            Whether Zipkin tracing is enabled.
        :param protocol_headers:
            Protocol-level headers to send with the request.
        """
        protocol_headers = protocol_headers or {}
        protocol_headers['as'] = 'thrift'
        return _ClientBase.__new__(
            cls, tchannel, hostport, service, trace, protocol_headers
        )

    new.__name__ = '__new__'
    methods = {'__new__': new}

    for method_name in method_names:
        methods[method_name] = generate_method(
            service_module, thrift_service_name, method_name
        )

    return type(thrift_service_name + 'Client', (_ClientBase,), methods)


def generate_method(service_module, service_name, method_name):
    """Generate a method for the given Thrift service.

    :param service_module:
        Thrift-generated service module
    :param service_name:
        Name of the Thrift service
    :param method_name:
        Method being called
    """
    assert service_module
    assert service_name
    assert method_name

    args_type = getattr(service_module, method_name + '_args')
    result_type = getattr(service_module, method_name + '_result', None)
    serializer = ThriftSerializer(result_type)
    # oneway not currently supported
    # TODO - write test for this
    if result_type is None:
        def not_supported(self, *args, **kwags):
            raise OneWayNotSupportedError(
                'TChannel+Thrift does not currently support oneway procedues'
            )
        return not_supported

    result_spec = result_type.thrift_spec
    # result_spec is a tuple of tuples in the form:
    #
    #   (fieldId, fieldType, fieldName, ...)
    #
    # Where "..." is other information we don't care about right now.
    #
    # result_spec will be empty if there is no return value or exception for
    # the method.
    #
    # Its first element, with field ID 0, contains the spec for the return
    # value. It is None if the result type is void but the method may still
    # throw exceptions.
    #
    # Elements after the first one are specs for the exceptions.

    endpoint = '%s::%s' % (service_name, method_name)

    @gen.coroutine
    def send(self, *args, **kwargs):
        params = inspect.getcallargs(
            getattr(service_module.Iface, method_name), self, *args, **kwargs
        )
        params.pop('self')  # self is already known

        # $methodName_args is the implicit struct containing the various
        # method parameters.
        call_args = args_type()
        for name, value in params.items():
            setattr(call_args, name, value)

        body = serializer.serialize_body(call_args)
        header = serializer.serialize_header({})

        # Glue for old API.
        if hasattr(self.tchannel, 'request'):
            response = yield self.tchannel.request(
                hostport=self.hostport, service=self.service
            ).send(
                arg1=endpoint,
                arg2=header,
                arg3=body,  # body
                headers=self.protocol_headers,
                traceflag=self.trace
            )
            body = yield response.get_body()
        else:
            response = yield self.tchannel.call(
                scheme=schemes.THRIFT,
                service=self.service,
                arg1=endpoint,
                arg2=header,
                arg3=body,
                hostport=self.hostport,
                # TODO: Need to handle these!
                # headers=self.protocol_headers,
                # traceflag=self.trace,
            )
            body = response.body

        call_result = serializer.deserialize_body(body)

        if not result_spec:
            # void return type and no exceptions allowed
            raise gen.Return(None)

        for exc_spec in result_spec[1:]:
            # May have failed with an exception
            exc = getattr(call_result, exc_spec[2])
            if exc is not None:
                raise exc

        if result_spec[0]:
            # Non-void return type. Return the result.
            success = getattr(call_result, result_spec[0][2])
            if success is not None:
                raise gen.Return(success)
        else:
            # No return type specified and no exceptions raised.
            raise gen.Return(None)

        # Expected a result but nothing was present in the object. Something
        # went wrong.
        raise Thrift.TApplicationException(
            Thrift.TApplicationException.MISSING_RESULT,
            '%s failed: did not receive a result as expected' % method_name
        )
        # TODO: We should probably throw a custom exception instead.

    send.__name__ = method_name
    return send



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#
# Autogenerated by Thrift Compiler (0.9.2)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:tornado,dynamic,slots,dynexc=VCRThriftError,dynbase=VCRThriftBase,dynimport=from tchannel.testing.vcr.thrift import *
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException

from tchannel.testing.vcr.thrift import *

class StatusCode(VCRThriftBase):
  SUCCESS = 0
  FAILURE = 1

  _VALUES_TO_NAMES = {
    0: "SUCCESS",
    1: "FAILURE",
  }

  _NAMES_TO_VALUES = {
    "SUCCESS": 0,
    "FAILURE": 1,
  }

class ArgScheme(VCRThriftBase):
  RAW = 0
  JSON = 1
  THRIFT = 2

  _VALUES_TO_NAMES = {
    0: "RAW",
    1: "JSON",
    2: "THRIFT",
  }

  _NAMES_TO_VALUES = {
    "RAW": 0,
    "JSON": 1,
    "THRIFT": 2,
  }


class TransportHeader(VCRThriftBase):
  """
  Attributes:
   - key
   - value
  """

  __slots__ = [ 
    'key',
    'value',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'key', None, None, ), # 1
    (2, TType.STRING, 'value', None, None, ), # 2
  )

  def __init__(self, key=None, value=None,):
    self.key = key
    self.value = value

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.key)
    value = (value * 31) ^ hash(self.value)
    return value


class Request(VCRThriftBase):
  """
  Attributes:
   - serviceName
   - endpoint
   - headers
   - body
   - hostPort: Remote hostport to which the request will be sent.

  If specified, the request will be made to this host only.
   - argScheme
   - transportHeaders
   - knownPeers: List of known peers of the TChannel at the time the request was made.

  This MUST be specified if hostPort wasn't. The request will be sent to
  a random peer from the list.
  """

  __slots__ = [ 
    'serviceName',
    'endpoint',
    'headers',
    'body',
    'hostPort',
    'argScheme',
    'transportHeaders',
    'knownPeers',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'serviceName', None, None, ), # 1
    (2, TType.STRING, 'endpoint', None, None, ), # 2
    (3, TType.STRING, 'headers', None, "", ), # 3
    (4, TType.STRING, 'body', None, None, ), # 4
    (5, TType.STRING, 'hostPort', None, "", ), # 5
    (6, TType.I32, 'argScheme', None,     0, ), # 6
    (7, TType.LIST, 'transportHeaders', (TType.STRUCT,(TransportHeader, TransportHeader.thrift_spec)), [
    ], ), # 7
    (8, TType.LIST, 'knownPeers', (TType.STRING,None), [
    ], ), # 8
  )

  def __init__(self, serviceName=None, endpoint=None, headers=thrift_spec[3][4], body=None, hostPort=thrift_spec[5][4], argScheme=thrift_spec[6][4], transportHeaders=thrift_spec[7][4], knownPeers=thrift_spec[8][4],):
    self.serviceName = serviceName
    self.endpoint = endpoint
    self.headers = headers
    self.body = body
    self.hostPort = hostPort
    self.argScheme = argScheme
    if transportHeaders is self.thrift_spec[7][4]:
      transportHeaders = [
    ]
    self.transportHeaders = transportHeaders
    if knownPeers is self.thrift_spec[8][4]:
      knownPeers = [
    ]
    self.knownPeers = knownPeers

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.serviceName)
    value = (value * 31) ^ hash(self.endpoint)
    value = (value * 31) ^ hash(self.headers)
    value = (value * 31) ^ hash(self.body)
    value = (value * 31) ^ hash(self.hostPort)
    value = (value * 31) ^ hash(self.argScheme)
    value = (value * 31) ^ hash(self.transportHeaders)
    value = (value * 31) ^ hash(self.knownPeers)
    return value


class Response(VCRThriftBase):
  """
  Attributes:
   - code
   - headers
   - body
  """

  __slots__ = [ 
    'code',
    'headers',
    'body',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'code', None, None, ), # 1
    (2, TType.STRING, 'headers', None, "", ), # 2
    (3, TType.STRING, 'body', None, None, ), # 3
  )

  def __init__(self, code=None, headers=thrift_spec[2][4], body=None,):
    self.code = code
    self.headers = headers
    self.body = body

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.code)
    value = (value * 31) ^ hash(self.headers)
    value = (value * 31) ^ hash(self.body)
    return value


class CannotRecordInteractionsError(VCRThriftError):
  """
  Raised when the record mode for a cassette prevents recording new
  interactions for it.

  Attributes:
   - message
  """

  __slots__ = [ 
    'message',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'message', None, None, ), # 1
  )

  def __init__(self, message=None,):
    self.message = message

  def __str__(self):
    return repr(self)

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.message)
    return value


class RemoteServiceError(VCRThriftError):
  """
  Raised when the remote service throws a protocol error.

  Attributes:
   - code
   - message
  """

  __slots__ = [ 
    'code',
    'message',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.BYTE, 'code', None, None, ), # 1
    (2, TType.STRING, 'message', None, None, ), # 2
  )

  def __init__(self, code=None, message=None,):
    self.code = code
    self.message = message

  def __str__(self):
    return repr(self)

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.code)
    value = (value * 31) ^ hash(self.message)
    return value


class NoPeersAvailableError(VCRThriftError):
  """
  Raised when both, hostPort and knownPeers were empty so the system couldn't
  make a request.

  Attributes:
   - message
  """

  __slots__ = [ 
    'message',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'message', None, None, ), # 1
  )

  def __init__(self, message=None,):
    self.message = message

  def __str__(self):
    return repr(self)

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.message)
    return value


class VCRServiceError(VCRThriftError):
  """
  A generic error for VCR exceptions not covered elsewhere.

  Attributes:
   - message
  """

  __slots__ = [ 
    'message',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'message', None, None, ), # 1
  )

  def __init__(self, message=None,):
    self.message = message

  def __str__(self):
    return repr(self)

  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.message)
    return value




# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from . import RAW


class RawArgScheme(object):
    """Semantic params and serialization for raw."""

    NAME = RAW

    def __init__(self, tchannel):
        self._tchannel = tchannel

    def __call__(
        self,
        service,
        endpoint,
        body=None,
        headers=None,
        timeout=None,
        retry_on=None,
        retry_limit=None,
        hostport=None,
        shard_key=None,
        trace=None,
    ):
        """Make a raw TChannel request.

        The request's headers and body are treated as raw bytes and not
        serialized/deserialized.

        The request's headers and body are treated as raw bytes and not
        serialized/deserialized.

        .. code-block: python

            from tchannel import TChannel

            tchannel = TChannel('my-service')

            resp = tchannel.raw(
                service='some-other-service',
                endpoint='get-all-the-crackers',
            )

        :param string service:
            Name of the service to call.
        :param string endpoint:
            Endpoint to call on service.
        :param string body:
            A raw body to provide to the endpoint.
        :param string headers:
            A raw headers block to provide to the endpoint.
        :param int timeout:
            How long to wait (in ms) before raising a ``TimeoutError`` - this
            defaults to ``tchannel.glossary.DEFAULT_TIMEOUT``.
        :param string retry_on:
            What events to retry on - valid values can be found in
            ``tchannel.retry``.
        :param string retry_limit:
            How many times to retry before
        :param string hostport:
            A 'host:port' value to use when making a request directly to a
            TChannel service, bypassing Hyperbahn.

        :rtype: Response
        """

        return self._tchannel.call(
            scheme=self.NAME,
            service=service,
            arg1=endpoint,
            arg2=headers,
            arg3=body,
            timeout=timeout,
            retry_on=retry_on,
            retry_limit=retry_limit,
            hostport=hostport,
            shard_key=shard_key,
            trace=trace,
        )

    def register(self, endpoint, **kwargs):

        # no args, eg - server.raw.register
        if callable(endpoint):
            handler = endpoint
            endpoint = None

        # args, eg - server.raw.register('bar')
        else:
            handler = None

        return self._tchannel.register(
            scheme=self.NAME,
            endpoint=endpoint,
            handler=handler,
            **kwargs
        )



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import random
import threading
from functools import wraps

from tornado import gen
from tornado.ioloop import IOLoop

from tchannel.errors import TChannelError
from tchannel.tornado import TChannel

from .proxy import VCRProxy


def wrap_uncaught(func=None, reraise=None):
    """Catches uncaught exceptions and raises VCRServiceErrors instead.

    :param reraise:
        Collection of exception clasess that should be re-raised as-is.
    """
    reraise = reraise or ()

    def decorator(f):

        @wraps(f)
        @gen.coroutine
        def new_f(*args, **kwargs):
            try:
                result = yield gen.maybe_future(f(*args, **kwargs))
            except Exception as e:
                if any(isinstance(e, cls) for cls in reraise):
                    # TODO maybe use traceback.format_exc to also send a
                    # traceback?
                    raise e
                raise VCRProxy.VCRServiceError(e.message)
            else:
                raise gen.Return(result)

        return new_f

    if func is not None:
        return decorator(func)
    else:
        return decorator


class VCRProxyService(object):

    def __init__(self, cassette, unpatch):
        """
        :param unpatch:
            A function returning a context manager which temporarily unpatches
            any monkey patched code so that a real request can be made.
        :param cassette:
            Cassette being played.
        """
        self.unpatch = unpatch
        self.cassette = cassette

        self.io_loop = None
        self.thread = None
        self.tchannel = None

        self._running = threading.Event()

    @wrap_uncaught(reraise=(
        VCRProxy.CannotRecordInteractionsError,
        VCRProxy.NoPeersAvailableError,
        VCRProxy.RemoteServiceError,
        VCRProxy.VCRServiceError,
    ))
    @gen.coroutine
    def send(self, request, response):
        cassette = self.cassette
        request = request.args.request

        # TODO decode requests and responses based on arg scheme into more
        # readable formats.

        # Because Thrift doesn't handle UTF-8 correctly right now
        request.serviceName = request.serviceName.decode('utf-8')
        request.endpoint = request.endpoint.decode('utf-8')

        # TODO do we care about hostport being the same?
        if cassette.can_replay(request):
            vcr_response = cassette.replay(request)
            raise gen.Return(vcr_response)

        if cassette.write_protected:
            raise VCRProxy.CannotRecordInteractionsError(
                'Could not find a matching response for request %s and the '
                'record mode %s prevents new interactions from being '
                'recorded. Your test may be performing an unexpected '
                'request.' % (str(request), cassette.record_mode)
            )

        peers = []
        if request.hostPort:
            peers = [request.hostPort]
        else:
            peers = request.knownPeers

        if not peers:
            raise VCRProxy.NoPeersAvailableError(
                'Both, hostPort and knownPeers were empty or unset. '
                'One of them must be specified and non-empty.'
            )

        arg_scheme = VCRProxy.ArgScheme.to_name(request.argScheme).lower()

        with self.unpatch():
            # TODO propagate other request and response parameters
            # TODO might make sense to tag all VCR requests with a protocol
            # header of some kind
            response_future = self.tchannel.request(
                service=request.serviceName,
                arg_scheme=arg_scheme,
                hostport=random.choice(peers),
            ).send(
                request.endpoint,
                request.headers,
                request.body,
                headers={h.key: h.value for h in request.transportHeaders},
            )

        # Don't actually yield while everything is unpatched.
        try:
            response = yield response_future
        except TChannelError as e:
            raise VCRProxy.RemoteServiceError(
                code=e.code,
                message=e.message,
            )
        response_headers = yield response.get_header()
        response_body = yield response.get_body()

        vcr_response = VCRProxy.Response(
            response.status_code,
            response_headers,
            response_body,
        )
        cassette.record(request, vcr_response)
        raise gen.Return(vcr_response)

    @property
    def hostport(self):
        return self.tchannel.hostport

    def _run(self):
        self.io_loop = IOLoop()
        self.io_loop.make_current()

        self.tchannel = TChannel('proxy-server')
        self.tchannel.register(VCRProxy, handler=self.send)

        self.tchannel.listen()
        self._running.set()
        self.io_loop.start()

    def start(self):
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

        self._running.wait(1)

    def stop(self):
        self.tchannel.close()
        self.tchannel = None

        self.io_loop.stop()
        self.io_loop = None

        self.thread.join(1)  # seconds
        self.thread = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()
